#include "PyTine.h"

PyTine* PyTine::m_instance = 0L;
string PyTine::address = "none";
int PyTine::timeout = 1000;

//Definition of the external prototype
//extern "C" void SystemSetCycleTimer(void);

/*
 * Constructor: finds the name server
 */
PyTine::PyTine()
{
	//SystemSetCycleTimer();
	PyTine::address = "none";
	PyTine::timeout = 1000; //set 1sec by default
	FindNameServerOnNetwork();
}

PyTine::~PyTine()
{
}

PyTine* PyTine::instance()
{
  return m_instance;
}


int PyTine::initConnection(char * adr, int timeout)
{
	PyTine::timeout = timeout;
	PyTine::address = adr;
	return 0;
}

/*
 * Set the PyTINE Address
 */
int PyTine::setAddress(char * address)
{
	PyTine::address = address;
	return 0;
}

/*
 * Set the PyTINE Timeout
 */
int PyTine::setTimeout(int timeout)
{
	PyTine::timeout = timeout;
	return 0;
}

/*
 * Returns to PyTINE TINE address
 */
const char * PyTine::getAddress()
{
	return PyTine::address.c_str();
}

/*
 * Returns to PyTINE the TINE timeout
 */
int PyTine::getTimeout()
{
	return PyTine::timeout;
}

/*
 * Executes the tget TINE command-line command
 */
int PyTine::tget(string address, string property, int size, int format, void * buf, int timeout) const
{
  int status = 0;
  DTYPE dout;

  // Build the TINE message
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = format;
  dout.dArrayLength = size;
  dout.data.vptr = buf;

  status = ExecLinkEx(address.c_str(),property.c_str(),&dout,NULL,CA_READ|CA_RETRY,timeout);

  return status;

}

/*
 * Executes the tget TINE command-line cmd for a STRUCT PROPERTY
 */
int PyTine::tgetStruct(string address, string property, int size, int format, void * buf, int timeout, string tag) const
{
  int status = 0;
  DTYPE dout;

  // Build the TINE message
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = format;
  dout.dArrayLength = size;
  dout.data.vptr = buf;

  strcpy(dout.dTag,tag.c_str());

  status = ExecLinkEx(address.c_str(),property.c_str(),&dout,NULL,CA_READ|CA_RETRY,timeout);

  return status;

}

/*
 * Executes the tget TINE command-line cmd (async)
 */
int PyTine::tgetAsync(string address, string property, int size, int format, void * buf, void (*func)(int cc,int st), int timeout, int mode) const
{
  int status = 0;
  DTYPE dout;

  // Build the TINE message
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = format;
  dout.dArrayLength = size;
  dout.data.vptr = buf;

  status = AttachLink(address.c_str(), property.c_str(), &dout, NULL, CA_READ|CA_RETRY,  timeout , func, mode);

  return status;

}

/*
 * Executes the tget TINE command-line cmd (async) (extended)
 */
int PyTine::tgetAsyncEx(string address, string property, int size, int format, void * buf, void (*func)(int cc,int st), int timeout, int mode, int identifier) const
{
  int status = 0;
  DTYPE dout;

  // Build the TINE message
  memset(&dout,0,sizeof(DTYPE));
  dout.dFormat = format;
  dout.dArrayLength = size;
  dout.data.vptr = buf;

  status = AttachLinkEx(address.c_str(), property.c_str(), &dout, NULL, CA_READ|CA_RETRY,  timeout , func, mode, identifier);

  return status;

}


/*
 * Executes the tput TINE command-line cmd
 */
int PyTine::tput(string address, string property, int size, int format, void * data) const
{
  int status = 0;
  DTYPE din;

  // Build the TINE message
  memset(&din,0,sizeof(DTYPE));
  din.dFormat = format;
  din.dArrayLength = size;
  din.data.vptr = data;

  status = ExecLinkEx(address.c_str(),property.c_str(),NULL,&din,CA_WRITE|CA_RETRY,PyTine::timeout);

  return status;

}


/*
 * Executes the tput TINE command-line cmd
 */
int PyTine::tputStruct(string address, string property, int size, int format, void * data, string tag) const
{
  int status = 0;
  DTYPE din;

  // Build the TINE message
  memset(&din,0,sizeof(DTYPE));
  din.dFormat = format;
  din.dArrayLength = size;
  din.data.vptr = data;

  strcpy(din.dTag,tag.c_str());

  status = ExecLinkEx(address.c_str(),property.c_str(),NULL,&din,CA_WRITE|CA_RETRY,PyTine::timeout);

  return status;

}


/*
 * Executes a put and get function
 */
int PyTine::tsendrecv(string address, string property, int size1, int format1, void * blob, int size2, int format2, void * data) const
{
	DTYPE din, dout;

	memset(&dout,0,sizeof(DTYPE));
	memset(&din,0,sizeof(DTYPE));

	din.dFormat = format1;
	din.dArrayLength = size1;
	din.data.vptr = data;

	dout.dFormat = format2;
	dout.dArrayLength = size2;
	dout.data.vptr = blob;

	return ExecLinkEx(address.c_str(),property.c_str(),&dout,&din,CA_READ|CA_WRITE|CA_RETRY,PyTine::timeout);

}


/*
 * Executes a put and get function
 */
int PyTine::tputget(string address, string property, int size, int format, void * blob, void * data) const
{
	DTYPE din, dout;

	memset(&dout,0,sizeof(DTYPE));
	memset(&din,0,sizeof(DTYPE));
	din.dFormat = format;
	din.dArrayLength = size;
	din.data.vptr = data;

	dout.dFormat = CF_DEFAULT;
	dout.dArrayLength = sizeof(data);
	dout.data.vptr = blob;

	return ExecLinkEx(address.c_str(),property.c_str(),&dout,&din,CA_READ|CA_WRITE|CA_RETRY,PyTine::timeout);

}

/*
 * Executes a put and get function
 */
int PyTine::tputgetStruct(string address, string property, int size, int format, void * blobstr, void * data) const
{
	DTYPE din, dout;

	memset(&dout,0,sizeof(DTYPE));
	memset(&din,0,sizeof(DTYPE));
	din.dFormat = format;
	din.dArrayLength = size;
	din.data.vptr = data;

	dout.dFormat = CF_NAME64DBLDBL;
	dout.dArrayLength = 512;
	dout.data.vptr = blobstr;

	return ExecLinkEx(address.c_str(),property.c_str(),&dout,&din,CA_READ|CA_WRITE|CA_RETRY,PyTine::timeout);

}



