#include "PyTine_py.h"

/*
 * TINE parsing function (used for pyput)
 */
BYTE *parseInputDataString(char *str,int *fmt,int *siz)
{
    BYTE *data=NULL;
    char *s, *c, *tok = " ,;\t\n";
    int i, n, len, base = 10;
    if (str == NULL || fmt == NULL || siz == NULL) return data;
    if (str[0] == '\"')
    {
      if ((c=strchr(&str[1],'\"')) == NULL) return data; /* badly formed */
      *c = 0;
      data = (BYTE *)&str[1];
      *siz = strlen((char *)data);
      *fmt = CF_TEXT;
    }
    else
    {
      len = strlen(str); n = 0;
      data = (BYTE *)calloc(len,64); /* worst case ? */
      for (s=strtok(str,tok); s != NULL; s = strtok(NULL,tok))
      {
        if (*fmt == CF_NULL)
        {
          len = strlen(s);
          for (i=0; i<len; i++) if (!isdigit(s[i])) break;
          if (i == len) /* all digits ! */
          {
            *fmt = CF_LONG;
            base = 10;
          }
          else if (strncmp(s,"0x",2) == 0 || strncmp(s,"0X",2) == 0)
          {
            *fmt = CF_LONG;
            base = 16;
          }
          else if (s[i] == '.')
          {
            *fmt = CF_DOUBLE;
          }
          else
          {
            if (len < 16) *fmt = CF_NAME16;
            else if (len < 32) *fmt = CF_NAME32;
            else if (len < 64) *fmt = CF_NAME64;
            else *fmt = CF_TEXT;
          }
        }
        switch (*fmt)
        {
          case CF_INT16:
            ((UINT16 *)data)[n] = (UINT16)strtol(s,NULL,base); break;
          case CF_INT32:
            ((UINT32 *)data)[n] = strtol(s,NULL,base); break;
          case CF_FLOAT:
            ((float *)data)[n] = (float)strtod(s,NULL); break;
          case CF_DOUBLE:
            ((double *)data)[n] = strtod(s,NULL); break;
          case CF_NAME16:
            strncpy(((NAME16 *)data)[n].name,s,16); break;
          case CF_NAME32:
            strncpy(((NAME32 *)data)[n].name,s,32); break;
          case CF_NAME64:
            strncpy(((NAME64 *)data)[n].name,s,64); break;
          case CF_TEXT:
            strncpy((char *)data,s,len);
            *siz = len;
            return data;
          default:
            break;
        }
        n++;
        if (*siz > 0 && n == *siz) break;
      }
      *siz = n;
    }
    return data;
}

/*
 * Parsing of a Python List to a char (TO BE CHECKED DECREF)
 */
char * pythonListToString(PyObject * pList)
{
	string data = "";
	int size = 0;
	PyObject * pString = NULL;

	//Extract the size
	size = PyList_Size(pList);

	//Extract the list an compose a char
	for (int i=0;i<size;i++)
	{
	  pString = PyList_GetItem(pList, i);
	  //Check the data type
	  if (!PyString_Check(pString))
	  {
	    //Py_DECREF(pList);
	  	PyErr_SetString(PyExc_TypeError,"pyput() wrong input type");
	  	return NULL;
	  }
	  data = data + PyString_AsString(pString) + " ";
	}
	//Py_DECREF(pList);
	return((char *) data.c_str());
}

PyObject * parseOutput(int fmt, int size)
{
	char s[65];
	int i = 0;
	int nlen = 0;
	PyObject * pList;

	switch(fmt)
	{
		case CF_LONG:
			for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("l", lval[i])); break;
		case CF_SHORT:
			for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("i", sval[i])); break;
		case CF_FLOAT:
			for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("f", fval[i])); break;
		case CF_DOUBLE:
			for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("d", dval[i])); break;
		case CF_TEXT:
			for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("s", cval[i])); break;
		case CF_NAME16:
		case CF_NAME32:
		case CF_NAME48:
		case CF_NAME64:
			nlen = fmtsizeof(fmt);
		  	for (i=0; i<size; i++)
		  	{
		  		strncpy(s,cval + i*nlen,nlen);
		  		PyList_SetItem(pList, i, Py_BuildValue("s", s));
		  	}
		  	break;
		 default:
			 PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported")); break;
	}
	return pList;
}

//Extract the tag of the property
char * extractTag(char * property)
{
	char * tag;
	char errstr[256];
	PrpQueryStruct * srvProps[10];
	srvProps[0] = new PrpQueryStruct;
	int * prpNum = new int;
	*prpNum = 1;
	int testresult = GetSystemPropertyInformation((char *) PyTine::instance()->getAddress(),(char *) property,srvProps,prpNum);
	if ( testresult == 0)
	{
		tag = srvProps[0]->prpTag;
	}
	else
	{
		cout << "Error " << GetLastLinkError(testresult,errstr) << endl;
		strcpy(tag, errstr);
	}

	return tag;
}

/*
 * Init the PyTine Connection
 */
static PyObject*  PyTine_init(PyObject*, PyObject *args)
{
  char* address;
  int timeout = 0;
  int sts = 0;
  //Get the input arguments from Python
  if (!PyArg_ParseTuple(args,"si", &address, &timeout))
  {
	PyErr_SetString(PyExc_TypeError, "invalid parameter");
	return NULL;
  }


  sts = PyTine::instance()->initConnection(address, timeout);

  if (sts == 0 ) initialized = true;
  else initialized = true;

  return Py_BuildValue("i", sts);

}

/*
 * Sets the TINE address
 */
static PyObject*  PyTine_setAddress(PyObject*, PyObject *args)
{
  char* address;
  int sts = 0;
  //Get the input arguments from Python
  if (!PyArg_ParseTuple(args,"s", &address))
  {
	PyErr_SetString(PyExc_TypeError, "invalid parameter");
	return NULL;
  }

  sts = PyTine::instance()->setAddress(address);

  return Py_BuildValue("i", sts);

}

/*
 * Returns the TINE address
 */
static PyObject* PyTine_getAddress(PyObject*, PyObject *args)
{
	const char * address = PyTine::instance()->getAddress();
	return Py_BuildValue("s", address);
}

/*
 * Returns the TINE timeout
 */
static PyObject* PyTine_getTimeout(PyObject*, PyObject *args)
{
	int timeout= PyTine::instance()->getTimeout();
	return Py_BuildValue("i", timeout);
}

/*
 * Sets the TINE address
 */
static PyObject*  PyTine_setTimeout(PyObject*, PyObject *args)
{
  int timeout = 1000;
  int sts = 0;
  //Get the input arguments from Python
  if (!PyArg_ParseTuple(args,"i", &timeout))
  {
	PyErr_SetString(PyExc_TypeError, "invalid parameter");
	return NULL;
  }

  sts = PyTine::instance()->setTimeout(timeout);

  return Py_BuildValue("i", sts);

}


/*
 * PyTget Method Main Implementation
 */
static PyObject* PyTine_tgetMain (const char* address,const char* property, const char* format, int size)
{
  //Arguments
  char errstr[256];
  char s[65];
  int i = 0;
  int nlen = 0;
  int fmt = 0;
  int sts = 0;

  //Check the size
  if (size > MAXIMUM_SIZE) size = MAXIMUM_SIZE;

  //Extract the format
  fmt = ftoi((char *) format);

  //Call the C code and returns a Python object
  sts = PyTine::instance()->tget(address, property, size, fmt, blob, PyTine::instance()->getTimeout());

  //No error in the TINE call
  if (sts == 0 )
  {
	  //Creates the PyThon list to return
	  PyObject *pList = PyList_New(size);
	  //assert(PyList_Check(pList));

	  switch(fmt)
	  {
	  	case CF_LONG:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("l", lval[i])); break;
	  	case CF_SHORT:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("i", sval[i])); break;
	  	case CF_FLOAT:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("f", fval[i])); break;
	  	case CF_DOUBLE:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("d", dval[i])); break;
	  	case CF_TEXT:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("s", cval[i])); break;
	  	case CF_NAME16:
	  	case CF_NAME32:
	  	case CF_NAME48:
	  	case CF_NAME64:
	  		nlen = fmtsizeof(fmt);
	  		for (i=0; i<size; i++)
	  		{
	  			strncpy(s,cval + i*nlen,nlen);
	  			PyList_SetItem(pList, i, Py_BuildValue("s", s));
	  		}
	  		break;
	  	case CF_STRUCT:
	  		//Resize the list to 1
	  		PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported, use pygetStruct")); break;
	  	default:
	  		//Resize the list to 1
	  		PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported")); break;
	  }
	  return pList;
	  //Free the memory
	  Py_DECREF(pList);
	  free(pList);
  }
  else
  {
	  //Creates the PyThon list to return
	  PyObject *pList = PyList_New(1);
	  assert(PyList_Check(pList));
	  //Return the TINE Error
	  PyList_SetItem(pList, 0, Py_BuildValue("s", GetLastLinkError(sts,errstr)));
	  return pList;
	  //Free the memory
	  Py_DECREF(pList);
	  free(pList);
  }
}

/*
 * PyTget Method Implementation
 * Get the parameters from python and calls PyTine_tgetMain()
 * const char* address, const char* property, const char* format, int size
 */
static PyObject* PyTine_tget (PyObject*, PyObject *args)
{
  const char* property;
  const char* format;
  int size = 0;
  //Get the input arguments from Python
  if (!PyArg_ParseTuple(args,"sis", &property, &size, &format))
  {
	PyErr_SetString(PyExc_TypeError, "invalid parameter");
	return NULL;
  }

  if (initialized)
    return  PyTine_tgetMain(PyTine::instance()->getAddress(), property, format, size);
  else
	return Py_BuildValue("s", "Not initialized");
}


PyObject * extractStructData(char * tag, int * initPos)
{
	char s[65];
	int nlen;
	int size = 0;
	int fmt = 0;
	int n,i,j;
	PyObject * pList;
	char * propStruct;
	float sizeStruct;
	float formatStruct;
	size_t pos;
	string tagStruct, tagProp, aux;
	short *sval = (short *)blob_aux;
	UINT32 *lval = (UINT32 *)blob_aux;
	float *fval = (float *)blob_aux;
	double *dval = blob_aux;
	char *cval = (char *)blob_aux;

	//Read the property structure
	pList =  PyTine_GetStructFormatMain(tag);

	//Extract the info of the structure
	n = PySequence_Length(pList);

	//Creates the PyThon list to return
	PyObject *pListStruct = PyList_New(n-1);

	int tmpInitPos = *initPos;

	PyObject *auxList;
	int counter=0;
	for (j=0; j<n-1; j++)
	{
		//Parse the tuple
		if (!PyArg_ParseTuple(PyList_GetItem(pList, j),"sff", &propStruct, &sizeStruct, &formatStruct))
		{
			PyErr_SetString(PyExc_TypeError, "invalid parameter");
			return NULL;
		}

		fmt = (int)formatStruct;
		size = (int) sizeStruct;
		nlen = fmtsizeof(fmt);

		int actualPos = (int) (tmpInitPos/nlen);

		int modulo = (tmpInitPos%nlen);

		//Adds the offset to the memory addresses
		if (modulo > 0)
		{
			//We move the addresses in bytes
			cval = cval + modulo;
			lval =  (UINT32 *) cval;
			sval = (short *) cval;
			fval = (float *) cval;
			dval = (double *) cval;
			tmpInitPos -= modulo;
		}

		//cout << "TmpPos= " << tmpInitPos << " InitPos= " << *initPos << " nlen= " << nlen << " ActualPos= " << actualPos << " Modulo= " << modulo << endl;

		//cout << "Init Position of the property " << propStruct << " " << *initPos << endl;
		//cout << "Size: " << sizeStruct << " nlen " << nlen << " Format " << formatStruct << endl;

		switch(fmt)
		{
			  	case CF_LONG:
			  		if (size > 1) // is an array
			  		{
			  			auxList = PyList_New(size);
			  			for (i=actualPos; i<size+actualPos; i++)
			  			{
			  				PyList_SetItem(auxList, counter, Py_BuildValue("l", lval[i]));
			  				counter++;
			  			}
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("O", auxList));;
			  		}
			  		else
			  		{
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("l", lval[actualPos]));
			  		}
			  		break;
			  	case CF_SHORT:
			  		if (size > 1) // is an array
			  		{
			  			auxList = PyList_New(size);
			  			for (i=actualPos; i<size+actualPos; i++)
			  			{
			  				PyList_SetItem(auxList, counter, Py_BuildValue("i", sval[i]));
			  				counter++;
			  			}
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("O", auxList));;
			  		}
			  		else
			  		{
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("i", sval[actualPos]));
			  		}
			  		break;
			  	case CF_FLOAT:
			  		if (size > 1) // is an array
			  		{
			  			auxList = PyList_New(size);
			  			for (i=actualPos; i<size+actualPos; i++)
			  			{
			  				PyList_SetItem(auxList, counter, Py_BuildValue("f", fval[i]));
			  				counter++;
			  			}
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("O", auxList));;
			  		}
			  		else
			  		{
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("f", fval[actualPos]));
			  		}
			  		break;
			  	case CF_DOUBLE:
			  		if (size > 1) // is an array
			  		{
			  			auxList = PyList_New(size);
			  			for (i=actualPos; i<size+actualPos; i++)
			  			{
			  				PyList_SetItem(auxList, counter, Py_BuildValue("d", dval[i]));
			  				counter++;
			  			}
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("O", auxList));;
			  		}
			  		else
			  		{
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("d", dval[actualPos]));
			  		}
			  		break;
			  	case CF_TEXT:
					strncpy(s,cval + actualPos,size);
					s[size] = '\0'; //terminate the char
			  		PyList_SetItem(pListStruct, j, Py_BuildValue("s", s)); break;
			  	case CF_NAME16:
			  	case CF_NAME32:
			  	case CF_NAME48:
			  	case CF_NAME64:
			  		nlen = fmtsizeof(fmt);
			  		if (size == 1)
			  		{
			  			strncpy(s,cval + actualPos,nlen);
			  			s[nlen] = '\0'; //terminate the char
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("s", s));
			  		}
			  		else
			  		{
			  			PyList_SetItem(pListStruct, j, Py_BuildValue("s", "arrays of NAMEXX in STRUCTS not supported"));
			  		}
			  		break;
			  	case CF_STRUCT:
					aux = propStruct;
					pos = aux.find("<");
					tagStruct = aux.substr(pos+1);
					pos = aux.find(">");
					tagStruct = tagStruct.substr(0,pos-1);
					aux = propStruct;
					pos = aux.find(">");
					tagProp = aux.substr(pos+1,aux.length());
					PyList_SetItem(pListStruct, j, extractStructData((char *)tagStruct.c_str(),initPos));
					break;
			  	default:
			  		//Resize the list to 1
			  		PyList_SetItem(pListStruct, j, Py_BuildValue("s", "type not supported")); break;
	  }

	  *initPos = (nlen*size) + *initPos;
	  tmpInitPos = (nlen*size) + tmpInitPos;

	} // of the for loop

	//free(pList);
	return pListStruct;
	//Free the memory
	//Py_DECREF(pListStruct);
	//free(pListStruct);
}

static PyObject* PyTine_tgetStruct (PyObject* self, PyObject *args)
{
  char* tag;
  char *property;
  char s[65];
  char errstr[256];
  int nlen;
  int size = 0;
  int sts = 0;
  int fmt = 0;
  int n,i,j;
  PyObject * pList;
  char * propStruct;
  float sizeStruct;
  float formatStruct;
  size_t pos;
  string tagStruct;
  string tagProp;
  string aux;
  short *sval = (short *)blob_aux;
  UINT32 *lval = (UINT32 *)blob_aux;
  float *fval = (float *)blob_aux;
  double *dval = blob_aux;
  char *cval = (char *)blob_aux;
  int increment = 0;

  if (initialized)
  {
	//Extract the PyThon arguments
	if (!PyArg_ParseTuple(args,"s", &property))
	{
		PyErr_SetString(PyExc_TypeError, "invalid parameter");
		return NULL;
	}

	tag = extractTag(property);

	//Read the property structure
	pList =  PyTine_GetStructFormatMain(tag);

	//Extract the info of the structure
	n = PySequence_Length(pList);

	//Extract the total data size of the structure (it is located in the last position)
	if (!PyArg_ParseTuple(PyList_GetItem(pList, n-1),"sff", &propStruct, &sizeStruct, &formatStruct))
	{
		PyErr_SetString(PyExc_TypeError, "invalid parameter");
		return NULL;
	}

	sts = PyTine::instance()->tgetStruct(PyTine::instance()->getAddress(), property, (int)sizeStruct, CF_STRUCT, blob, PyTine::instance()->getTimeout(), tag);

	//Copy the content in and auxiliary blob
	for(i=0;i<MAXIMUM_SIZE;i++)
	{
		blob_aux[i] = blob[i];
	}

    //No error in the TINE call
	if (sts == 0 )
	{
	  //Creates the PyThon list to return
	  PyObject *pListStruct = PyList_New(n-1);
	  //assert(PyList_Check(pListStruct));

	  int globalPos = 0;
	  int tmpPos = 0;

	  PyObject *auxList;
	  int counter=0;
	  for (j=0; j<n-1; j++)
	  {
		  //Parse the tuple
		  if (!PyArg_ParseTuple(PyList_GetItem(pList, j),"sff", &propStruct, &sizeStruct, &formatStruct))
		  {
			  PyErr_SetString(PyExc_TypeError, "invalid parameter");
		  	  return NULL;
		  }

		  fmt = (int)formatStruct;
		  size = (int) sizeStruct;
		  nlen = fmtsizeof(fmt);

		  //tmpPos = globalPos;

		  int actualPos = (tmpPos/nlen);

		  int modulo = (tmpPos%nlen);

		  //cout << "GlobalPos " << globalPos << " TmpPos " << tmpPos << " Actual Pos= " << actualPos << " Modulo= " << modulo << endl;

		  //Adds the offset to the memory addresses
		  if (modulo > 0)
		  {
			//We move the addresses in bytes
			cval = cval + modulo;
			lval =  (UINT32 *) cval;
			sval = (short *) cval;
			fval = (float *) cval;
			dval = (double *) cval;
			tmpPos -= modulo;
		  }


		  switch(fmt)
		  {
		  	case CF_LONG:
		  		if (size > 1) // is an array
		  		{
		  			auxList = PyList_New(size);
		  			for (i=actualPos; i<size+actualPos; i++)
		  			{
		  				PyList_SetItem(auxList, counter, Py_BuildValue("l", lval[i]));
		  				counter++;
		  			}
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("O", auxList));;
		  		}
		  		else
		  		{
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("l", lval[actualPos]));
		  		}
		  		break;
		  	case CF_SHORT:
		  		if (size > 1) // is an array
		  		{
		  			auxList = PyList_New(size);
		  			for (i=actualPos; i<size+actualPos; i++)
		  			{
		  				PyList_SetItem(auxList, counter, Py_BuildValue("i", sval[i]));
		  				counter++;
		  			}
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("O", auxList));;
		  		}
		  		else
		  		{
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("i", sval[actualPos]));
		  		}
		  		break;
		  	case CF_FLOAT:
		  		if (size > 1) // is an array
		  		{
		  			auxList = PyList_New(size);
		  			for (i=actualPos; i<size+actualPos; i++)
		  			{
		  				PyList_SetItem(auxList, counter, Py_BuildValue("f", fval[i]));
		  				counter++;
		  			}
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("O", auxList));;
		  		}
		  		else
		  		{
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("f", fval[actualPos]));
		  		}
		  		break;
		  	case CF_DOUBLE:
		  		if (size > 1) // is an array
		  		{
		  			auxList = PyList_New(size);
		  			for (i=actualPos; i<size+actualPos; i++)
		  			{
		  				PyList_SetItem(auxList, counter, Py_BuildValue("d", dval[i]));
		  				counter++;
		  			}
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("O", auxList));;
		  		}
		  		else
		  		{
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("d", dval[actualPos]));
		  		}
		  		break;
		  	case CF_TEXT:
		  		strncpy(s,cval + actualPos,size);
		  		s[size] = '\0'; //terminate the char
		  		PyList_SetItem(pListStruct, j, Py_BuildValue("s", s)); break;
		  	case CF_NAME16:
		  	case CF_NAME32:
		  	case CF_NAME48:
		  	case CF_NAME64:
		  		nlen = fmtsizeof(fmt);
		  		if (size == 1)
		  		{
		  			strncpy(s,cval + actualPos,nlen);
		  			s[size] = '\0'; //terminate the char
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("s", s));
		  		}
		  		else
		  		{
		  			PyList_SetItem(pListStruct, j, Py_BuildValue("s", "arrays of NAMEXX in STRUCTS not supported"));
		  		}
		  		break;
		  	case CF_STRUCT:
				aux = propStruct;
				pos = aux.find("<");
				tagStruct = aux.substr(pos+1);
				pos = aux.find(">");
				tagStruct = tagStruct.substr(0,pos-1);
				aux = propStruct;
				pos = aux.find(">");
				tagProp = aux.substr(pos+1,aux.length());
				increment = globalPos;
				PyList_SetItem(pListStruct, j, extractStructData((char *)tagStruct.c_str(),&globalPos));
				increment = globalPos - increment;
				break;
		  	default:
		  		//Resize the list to 1
		  		PyList_SetItem(pListStruct, j, Py_BuildValue("s", "type not supported")); break;
		  }

		  //The size of the structure is calculated by reference at the extractStructData
		  if (fmt != CF_STRUCT)
		  {
			  globalPos += nlen * size ;
			  tmpPos += nlen * size;
		  }
		  else
		  {
			  tmpPos += increment;
		  }

	  } // of the for loop
	  //free(pList);
	  return pListStruct;
	  //Free the memory
	  //Py_DECREF(pListStruct);
	  //free(pListStruct);
    }
	else
	{
		//Creates the PyThon list to return
		PyObject *pListError = PyList_New(1);
		assert(PyList_Check(pListError));
		//Return the TINE Error
		PyList_SetItem(pListError, 0, Py_BuildValue("s", GetLastLinkError(sts,errstr)));
		return pListError;
		//Free the memory
		Py_DECREF(pListError);
		free(pListError);
	}
  }
  else return Py_BuildValue("s", "Not initialized");

  return Py_BuildValue("s", "Not initialized");

}

/*
 * PyTgetExtended Method
 */
static PyObject* PyTine_tgetEx (PyObject*, PyObject *args)
{
  //Arguments
  const char* address;
  const char* property;
  const char* format;
  int size = 0;

  //Get the input arguments from Python
  if (!PyArg_ParseTuple(args,"ssis", &address, &property, &size, &format))
  {
	PyErr_SetString(PyExc_TypeError, "invalid parameter");
	return NULL;
  }

  return  PyTine_tgetMain(address, property, format, size);
}

static PyObject* PyTine_tsendrecvMain (const char* address,const char* property, const char* format1, int size1, const char* format2, int size2, char * datastr)
{
  //Arguments
  char errstr[256];
  char s[65];
  int i = 0;
  int nlen = 0;
  int fmt1 = 0;
  int fmt2 = 0;
  int sts = 0;
  NAME64DBLDBL inputStruct;

  //Check the size
  if (size1 > MAXIMUM_SIZE) size1 = MAXIMUM_SIZE;
  if (size2 > MAXIMUM_SIZE) size2 = MAXIMUM_SIZE;

  //Extract the format
  fmt1 = ftoi((char *) format1);
  fmt2 = ftoi((char *) format2);

  BYTE *data = NULL;
  data = parseInputDataString(datastr,&fmt1,&size1);

  //Call the C code and returns a Python object
  sts = PyTine::instance()->tsendrecv(address,property,size1,fmt1,blob,size2,fmt2,datastr);

  //No error in the TINE call
  if (sts == 0 )
  {

	  //Get the format and the size of the output from the last TINE call
	  //fmt = GetCompletionDataType(-1);
	  //size = GetCompletionDataSize(-1);

	  //Creates the PyThon list to return
	  PyObject *pList = PyList_New(size2);
	  //assert(PyList_Check(pList));
	  switch(fmt2)
	 	  {
	 	  	case CF_LONG:
	 	  		for (i=0; i<size2; i++) PyList_SetItem(pList, i, Py_BuildValue("l", lval[i])); break;
	 	  	case CF_SHORT:
	 	  		for (i=0; i<size2; i++) PyList_SetItem(pList, i, Py_BuildValue("i", sval[i])); break;
	 	  	case CF_FLOAT:
	 	  		for (i=0; i<size2; i++) PyList_SetItem(pList, i, Py_BuildValue("f", fval[i])); break;
	 	  	case CF_DOUBLE:
	 	  		for (i=0; i<size2; i++) PyList_SetItem(pList, i, Py_BuildValue("d", dval[i])); break;
	 	  	case CF_TEXT:
	 	  		for (i=0; i<size2; i++) PyList_SetItem(pList, i, Py_BuildValue("s", cval[i])); break;
	 	  	case CF_NAME16:
	 	  	case CF_NAME32:
	 	  	case CF_NAME48:
	 	  	case CF_NAME64:
	 	  		nlen = fmtsizeof(fmt2);
	 	  		for (i=0; i<size2; i++)
	 	  		{
	 	  			strncpy(s,cval + i*nlen,nlen);
	 	  			PyList_SetItem(pList, i, Py_BuildValue("s", s));
	 	  		}
	 	  		break;
	 	  	case CF_NAME64DBLDBL:
	 	  		nlen = fmtsizeof(CF_NAME64DBLDBL);
	 	  		for (i=0; i<size2; i++)
	 	  		{
	 	  			//Extract the NAME64
	 	  			inputStruct = structFormat[i];
	 	  			PyList_SetItem(pList, i, Py_BuildValue("sff", inputStruct.name, inputStruct.d1val, inputStruct.d2val));
	 	  		}
	 	  		break;
	 	  	default:
	 	  		PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported")); break;
	 	  }
	 	  return pList;
	 	  //Free the memory
	 	  //Py_DECREF(pList);
	 	  //free(pList);
	   }
	   else
	   {
	 	  //Creates the PyThon list to return
	 	  PyObject *pList = PyList_New(1);
	 	  assert(PyList_Check(pList));
	 	  //Return the TINE Error
	 	  PyList_SetItem(pList, 0, Py_BuildValue("s", GetLastLinkError(sts,errstr)));
	 	  return pList;
	 	  //Free the memory
	 	  //Py_DECREF(pList);
	 	  //free(pList);
	   }
}

/*
 * Pytputget Method Implementation
 */
static PyObject* PyTine_tsendrecv (PyObject* pSelf, PyObject *args)
{
  //Arguments
  const char* property;
  const char* format1;
  const char* format2;
  int size1 = 0;
  int size2 = 0;
  int listSize = 0;
  char* datastr;
  PyObject *pList = NULL;

  //Extract the PyThon arguments
  if (!PyArg_ParseTuple(args,"ssisiO", &property, &format1, &size1, &format2, &size2, &pList))
  {
      PyErr_SetString(PyExc_TypeError, "invalid parameter");
    	return NULL;
  }

  //Check that the fifth parameter is a list
  if(!PyList_Check(pList))
  {
    PyErr_SetString(PyExc_TypeError,"pyput() fifth parameter must be a list");
    return NULL;
  }

  listSize = PyList_Size(pList);

  //datastr = pythonListToString(pList);
  string data = "";
  PyObject * pString = NULL;

  //Extract the list an compose a char
  for (int i=0;i<listSize;i++)
  {
    	pString = PyList_GetItem(pList, i);
  	//Check the data type
  	if (!PyString_Check(pString))
   	{
   	  PyErr_SetString(PyExc_TypeError,"pyput() wrong input type");
   	  return NULL;
   	}
   	data = data + PyString_AsString(pString) + " ";
  }
  datastr = (char *) data.c_str();

  return PyTine_tsendrecvMain(PyTine::instance()->getAddress(),property, format1, size1, format2, size2, datastr);
}

/*
 * Pytputget Extended Method Implementation
 */
static PyObject* PyTine_tsendrecvEx (PyObject* pSelf, PyObject *args)
{
  //Arguments
  const char* address;
  const char* property;
  const char* format1;
  const char* format2;
  char* datastr;
  int listSize = 0;
  int size1 = 0;
  int size2 = 0;
  PyObject *pList = NULL;

  //Extract the PyThon arguments
  if (!PyArg_ParseTuple(args,"sssisiO", &address, &property, &format1, &size1, &format2, &size2, &pList))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
    return NULL;
  }

  //Check that the fifth parameter is a list
  if(!PyList_Check(pList))
  {
    PyErr_SetString(PyExc_TypeError,"pyput() fifth parameter must be a list");
    return NULL;
  }

  listSize = PyList_Size(pList);

  //datastr = pythonListToString(pList);
  string data = "";
  PyObject * pString = NULL;

  //Extract the list an compose a char
  for (int i=0;i<listSize;i++)
  {
	pString = PyList_GetItem(pList, i);
  	//Check the data type
  	if (!PyString_Check(pString))
  	{
  	  PyErr_SetString(PyExc_TypeError,"pyput() wrong input type");
  	  return NULL;
  	}
  	data = data + PyString_AsString(pString) + " ";
  }
  datastr = (char *) data.c_str();

  return PyTine_tsendrecvMain(address,property, format1, size1, format2, size2, datastr);

}




/*
 * Pytput Method Main Implementation
 */
static PyObject* PyTine_tputMain (const char* address, const char* property, const char* format, int size, char * datastr)
{
  //Arguments
  char errstr[256];
  int fmt = 0;
  int sts = 0;
  PyObject *pList;

  //Check the size
  if (size > MAXIMUM_SIZE) size = MAXIMUM_SIZE;

  //Extract the format
  fmt = ftoi((char *) format);

  //Call the C code and returns a Python object
  BYTE *data = NULL;
  data = parseInputDataString(datastr,&fmt,&size);
  sts = PyTine::instance()->tput(address, property, size, fmt, data);

  if (sts == 0 ) return Py_BuildValue("s", "Success");
  else return Py_BuildValue("s", GetLastLinkError(sts,errstr));
  Py_DECREF(pList);

}

/*
 * Pytput Method Main Implementation
 */
static PyObject* PyTine_tput(PyObject*, PyObject *args)
{
  //Arguments
  const char* property;
  const char* format;
  char* datastr;
  int size = 0;
  PyObject *pList = NULL;
  PyObject *pString = NULL;

  if (!PyArg_ParseTuple(args,"ssO", &property, &format, &pList))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  //Check that the fifth parameter is a list
  if(!PyList_Check(pList))
  {
	  PyErr_SetString(PyExc_TypeError,"pyput() third parameter must be a list");
      return NULL;
  }

  // Extract the List size
  size = PyList_Size(pList);

  string data = "";
  string tempString = "";
  long auxLong;
  double auxDouble;

  //Extract the list an compose a char
  for (int i=0;i<size;i++)
  {
	  tempString = "";
	  stringstream str;
	  pString = PyList_GetItem(pList, i);
	  //Check the data type
	  if (!PyString_Check(pString))
	  {
		switch (ftoi((char *) format))
		{
			case CF_SHORT:
			case CF_LONG:
				auxLong = PyLong_AsLong(pString);
				str << auxLong;
				tempString = str.str();
				break;
			case CF_DOUBLE:
			case CF_FLOAT:
				auxDouble = PyFloat_AsDouble(pString);
				str << auxDouble;
				tempString = str.str();
				break;
			default:
				Py_DECREF(pList);
	  			PyErr_SetString(PyExc_TypeError,"pyput() wrong input type");
	  			return NULL;
		}
	  }
	  else tempString = PyString_AsString(pString);
	  data = data +  tempString + " ";
  }
  Py_DECREF(pList);
  datastr = (char *) data.c_str();


  if (initialized)
    return PyTine_tputMain(PyTine::instance()->getAddress(),property,format,size, datastr);
  else
    return Py_BuildValue("s", "Not initialized");
}

/*
 * Pytput Struct Method Main Implementation
 * Note: the input data has to be a list of strings
 * The only limitation I see is that I can not put directly a structure that I get
 * has to be a conversion to string in between?
 */
static PyObject* PyTine_tputStruct(PyObject* self, PyObject *args)
{
  //Arguments
  char errstr[256];
  char* property;
  int size = 0;
  int fmt, sts, nlen, offset;
  int n = 0;
  int totalSize = 0 ;
  PyObject *pString = NULL;
  PyObject *pList = NULL;
  PyObject *plistStruct = NULL;
  char * propStruct;
  float sizeStruct,formatStruct;
  BYTE *data = NULL;
  BYTE *finalData = NULL;
  string aux, tagStruct, tagProp;
  char * tag;

  if (!PyArg_ParseTuple(args,"sO", &property, &pList))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  //Check that second parameter was a list
  if(!PyList_Check(pList))
  {
    PyErr_SetString(PyExc_TypeError,"pyputStruct() second parameter must be a list");
    return NULL;
  }

  //Read the property structure
  tag = extractTag(property);
  plistStruct =  PyTine_GetStructFormatMain (tag);
  //Extract the length of the structure
  n = PySequence_Length(plistStruct) - 1 ;
  size = PyList_Size(pList);

  if (n != size)
  {
	  PyErr_SetString(PyExc_TypeError, "invalid list parameter length");
	  return NULL;
  }

  //Extract the total data size of the structure (it is located in the last position)
  if (!PyArg_ParseTuple(PyList_GetItem(plistStruct, n),"sff", &propStruct, &sizeStruct, &formatStruct))
  {
  		PyErr_SetString(PyExc_TypeError, "invalid parameter");
  		return NULL;
  }
  totalSize = (int) sizeStruct;

  //cout << propStruct << " " << sizeStruct << " " << formatStruct << endl;


  //Build a memory block with the data to send to TINE
  char * tempData;
  offset = 0;
  finalData = (BYTE *) malloc(totalSize * sizeof(char));

  for (int j=0; j<n; j++)
  {
    //Extract the Info of the Structure
  	if (!PyArg_ParseTuple(PyList_GetItem(plistStruct, j),"sff", &propStruct, &sizeStruct, &formatStruct))
  	{
  	  PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	  return NULL;
  	}

  	fmt = (int)formatStruct;
  	size = (int) sizeStruct;

  	//cout << "Format Prop " << j << " -> " << fmt << " Size Prop " << j << " -> "  << size  << endl;

	//int pos = 0;
	if (fmt == CF_STRUCT)
	{
		//cout << propStruct << endl;
		//string aux1(propStruct);
		/*pos = aux1.find("<");
		tagStruct = aux1.substr(pos+1);
		cout << tagStruct << endl;
		pos = aux1.find(">");
		tagStruct = tagStruct.substr(0,pos-1);
		cout << "I am a struct " << tagStruct << endl;*/
	}
	else
	{
		//Extract the Data to copy
		/*pString = PyList_GetItem(pList, j);

		//Check the data type
		if (!PyString_Check(pString))
		{
			Py_DECREF(pList);
			PyErr_SetString(PyExc_TypeError,"pyputStruct wrong input type");
			return NULL;
		}
		Py_DECREF(pList);
		tempData = (char *) PyString_AsString(pString);*/

		//cout << "Input data " << tempData << endl;

		string tempString = "";
		long auxLong;
		double auxDouble;
		stringstream str;
		pString = PyList_GetItem(pList, j);
		//Check the data type
		if (!PyString_Check(pString))
		{
		  switch (fmt)
		  {
			case CF_SHORT:
			case CF_LONG:
			  auxLong = PyLong_AsLong(pString);
			  str << auxLong;
			  tempString = str.str();
			  break;
			case CF_DOUBLE:
			case CF_FLOAT:
			  auxDouble = PyFloat_AsDouble(pString);
			  str << auxDouble;
			  tempString = str.str();
			  break;
			default:
			  Py_DECREF(pList);
		  	  PyErr_SetString(PyExc_TypeError,"pyput() wrong input type");
		  	  return NULL;
		  }
		}
		else tempString = PyString_AsString(pString);
		tempData = (char *) tempString.c_str();

		if (fmt != CF_TEXT)
			data = parseInputDataString(tempData,&fmt,&size);
		else
			memcpy(data, tempData, size);
		nlen = fmtsizeof(fmt);
		memcpy(finalData+offset, data, nlen*size);
		offset += nlen*size;
	}

  } //of the for loop

  if (initialized)
  {
    sts = PyTine::instance()->tputStruct(PyTine::instance()->getAddress(), property, totalSize, CF_STRUCT, finalData, property);
	if (sts == 0 ) return Py_BuildValue("s", "Success");
	else return Py_BuildValue("s", GetLastLinkError(sts,errstr));
  }
  else return Py_BuildValue("s", "PyTine not initialized");
  return Py_BuildValue("s", "Testing");

}


/*
 * Pytput Struct Method Main Implementation
 * Note: the input data has to be a list of strings
 * The only limitation I see is that I can not put directly a structure that I get
 * has to be a conversion to string in between
 */
static PyObject* PyTine_tputArrayStruct(PyObject* self, PyObject *args)
{
  //Arguments
  char errstr[256];
  char* property;
  int size = 0;
  int listSize = 0;
  int fmt, sts, nlen, offset;
  int n = 0;
  int totalSize = 0 ;
  PyObject *pString = NULL;
  PyObject *pList = NULL;
  PyObject *plistStruct = NULL;
  char * propStruct;
  float sizeStruct,formatStruct;
  BYTE *data = NULL;
  BYTE *finalData = NULL;
  char * tag;

  if (!PyArg_ParseTuple(args,"sO", &property, &pList))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  //Check that second parameter was a list
  if(!PyList_Check(pList))
  {
    PyErr_SetString(PyExc_TypeError,"pyputStruct() second parameter must be a list");
    return NULL;
  }

  //Extract the tag
  tag = extractTag(property);

  //Read the property structure
  plistStruct =  PyTine_GetStructFormatMain (tag);
  //Extract the length of the structure
  n = PySequence_Length(plistStruct) - 1 ;
  listSize = PyList_Size(pList);

  /*if (n != size)
  {
	  PyErr_SetString(PyExc_TypeError, "invalid list parameter length");
	  return NULL;
  }*/

  //Extract the total data size of the structure (it is located in the last position)
  if (!PyArg_ParseTuple(PyList_GetItem(plistStruct, n),"sff", &propStruct, &sizeStruct, &formatStruct))
  {
  		PyErr_SetString(PyExc_TypeError, "invalid parameter");
  		return NULL;
  }
  totalSize = (int) sizeStruct * listSize;

  //Build a memory block with the data to send to TINE
  char * tempData;
  offset = 0;
  finalData = (BYTE *) malloc(totalSize * sizeof(char));

  for (int aux=0; aux< listSize;aux++)
  {
	  for (int j=0; j<n; j++)
	  {
		  //Extract the Info of the parameters of the Structure
		  if (!PyArg_ParseTuple(PyList_GetItem(plistStruct, j),"sff", &propStruct, &sizeStruct, &formatStruct))
		  {
			  PyErr_SetString(PyExc_TypeError, "invalid parameter");
			  return NULL;
		  }

		  fmt = (int)formatStruct;
		  size = (int) sizeStruct;

		  //Extract the Data to copy
		  pString = PyList_GetItem(PyList_GetItem(pList, aux), j);
		  //Check the data type
		  if (!PyString_Check(pString))
		  {
			  Py_DECREF(pList);
			  PyErr_SetString(PyExc_TypeError,"pyputStruct wrong input type");
			  return NULL;
		  }
		  //Py_DECREF(pList);
		  tempData = (char *) PyString_AsString(pString);
		  if (fmt != CF_TEXT)
			  data = parseInputDataString(tempData,&fmt,&size);
		  else
			  memcpy(data, tempData, size);
		  nlen = fmtsizeof(fmt);
		  memcpy(finalData+offset, data, nlen*size);
		  offset += nlen*size;
	  }
  }
  if (initialized)
  {
    sts = PyTine::instance()->tputStruct(PyTine::instance()->getAddress(), property, totalSize, CF_STRUCT, finalData, property);
	if (sts == 0 ) return Py_BuildValue("s", "Success");
	else return Py_BuildValue("s", GetLastLinkError(sts,errstr));
  }
  else return Py_BuildValue("s", "PyTine not initialized");
  return Py_BuildValue("s", "Testing");

}



/*
 * Pytput Method Implementation
 */
static PyObject* PyTine_tputEx (PyObject* pSelf, PyObject *args)
{
  //Arguments
  const char* address;
  const char* property;
  const char* format;
  char* datastr;
  PyObject *pString = NULL;
  int size = 0;
  PyObject * pList = NULL;

  if (!PyArg_ParseTuple(args,"sssO", &address, &property, &format, &pList))
  {
	PyErr_SetString(PyExc_TypeError, "invalid parameter");
    return NULL;
  }

  //Check that the fifth parameter is a list
  if(!PyList_Check(pList))
  {
	PyErr_SetString(PyExc_TypeError,"pyput() fourth parameter must be a list");
    return NULL;
  }

  // Extract the List size
  size = PyList_Size(pList);

  string data = "";
  string tempString = "";

  long auxLong;
  double auxDouble;

  //Extract the list an compose a char
  for (int i=0;i<size;i++)
  {
	  tempString = "";
	  stringstream str;
	  pString = PyList_GetItem(pList, i);
	  //Check the data type
	  if (!PyString_Check(pString))
	  {
		switch (ftoi((char *) format))
		{
			case CF_SHORT:
			case CF_LONG:
				auxLong = PyLong_AsLong(pString);
				str << auxLong;
				tempString = str.str();
				break;
			case CF_DOUBLE:
			case CF_FLOAT:
				auxDouble = PyFloat_AsDouble(pString);
				str << auxDouble;
				tempString = str.str();
				break;
			default:
				Py_DECREF(pList);
	  			PyErr_SetString(PyExc_TypeError,"pyput() wrong input type");
	  			return NULL;
		}
	  }
	  else tempString = PyString_AsString(pString);
	  data = data +  tempString + " ";
  }
  Py_DECREF(pList);
  datastr = (char *) data.c_str();

  //datastr = pythonListToString(pList);

  return PyTine_tputMain(address,property,format,size, datastr);
}


/*
 * PyTget Method Main Implementation
 */
static PyObject* PyTine_tputgetMain (const char* address,const char* property, const char* format, int size, char * datastr)
{
  //Arguments
  char errstr[256];
  char s[65];
  int i = 0;
  int nlen = 0;
  int fmt = 0;
  int sts = 0;
  NAME64DBLDBL inputStruct;

  //Check the size
  if (size > MAXIMUM_SIZE) size = MAXIMUM_SIZE;

  //Extract the format
  fmt = ftoi((char *) format);

  BYTE *data = NULL;
  data = parseInputDataString(datastr,&fmt,&size);

  //Call the C code and returns a Python object
  sts = PyTine::instance()->tputget(address,property,size,fmt,blob,datastr);

  //No error in the TINE call
  if (sts == 0 )
  {

	  //Get the format and the size of the output from the last TINE call
	  fmt = GetCompletionDataType(-1);
	  size = GetCompletionDataSize(-1);

	  //Creates the PyThon list to return
	  PyObject *pList = PyList_New(size);
	  assert(PyList_Check(pList));

	  switch(fmt)
	  {
	  	case CF_LONG:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("l", lval[i])); break;
	  	case CF_SHORT:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("i", sval[i])); break;
	  	case CF_FLOAT:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("f", fval[i])); break;
	  	case CF_DOUBLE:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("d", dval[i])); break;
	  	case CF_TEXT:
	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("s", cval[i])); break;
	  	case CF_NAME16:
	  	case CF_NAME32:
	  	case CF_NAME48:
	  	case CF_NAME64:
	  		nlen = fmtsizeof(fmt);
	  		for (i=0; i<size; i++)
	  		{
	  			strncpy(s,cval + i*nlen,nlen);
	  			PyList_SetItem(pList, i, Py_BuildValue("s", s));
	  		}
	  		break;
	  	case CF_NAME64DBLDBL:
	  		nlen = fmtsizeof(CF_NAME64DBLDBL);
	  		for (i=0; i<size; i++)
	  		{
	  			//Extract the NAME64
	  			inputStruct = structFormat[i];
	  			PyList_SetItem(pList, i, Py_BuildValue("sff", inputStruct.name, inputStruct.d1val, inputStruct.d2val));
	  		}
	  		break;
	  	default:
	  		PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported")); break;
	  }
	  return pList;
	  //Free the memory
	  //Py_DECREF(pList);
	  //free(pList);
  }
  else
  {
	  //Creates the PyThon list to return
	  PyObject *pList = PyList_New(1);
	  assert(PyList_Check(pList));
	  //Return the TINE Error
	  PyList_SetItem(pList, 0, Py_BuildValue("s", GetLastLinkError(sts,errstr)));
	  return pList;
	  //Free the memory
	  //Py_DECREF(pList);
	  //free(pList);
  }
}

/*
 * Pytputget Method Implementation
 */
static PyObject* PyTine_tputget (PyObject* pSelf, PyObject *args)
{
  //Arguments
  const char* property;
  const char* format;
  char* datastr;
  int size = 0;
  PyObject *pList = NULL;

  //Extract the PyThon arguments
  if (!PyArg_ParseTuple(args,"ssO", &property, &format, &pList))
  {
      PyErr_SetString(PyExc_TypeError, "invalid parameter");
    	return NULL;
  }

  //Check that the fifth parameter is a list
  if(!PyList_Check(pList))
  {
    PyErr_SetString(PyExc_TypeError,"pyput() fifth parameter must be a list");
    return NULL;
  }

  size = PyList_Size(pList);

  //datastr = pythonListToString(pList);
  string data = "";
  PyObject * pString = NULL;
  string tempString = "";
  stringstream str;
  long auxLong;
  double auxDouble;

  //Extract the list an compose a char
  for (int i=0;i<size;i++)
  {
    pString = PyList_GetItem(pList, i);
  	//Check the data type
  	if (!PyString_Check(pString))
   	{
   	  	switch (ftoi((char *) format))
		{
			case CF_SHORT:
			case CF_LONG:
				auxLong = PyLong_AsLong(pString);
				str << auxLong;
				str >> tempString;
				break;
			case CF_DOUBLE:
			case CF_FLOAT:
				auxDouble = PyFloat_AsDouble(pString);
				str << auxDouble;
				str >> tempString;
				break;
			default:
				Py_DECREF(pList);
	  			PyErr_SetString(PyExc_TypeError,"pyput() wrong input type");
	  			return NULL;
		}
	}
	else tempString = PyString_AsString(pString);
   	data = data + tempString + " ";
  }
  datastr = (char *) data.c_str();
  //datastr = pythonListToString(pList);

  return PyTine_tputgetMain(PyTine::instance()->getAddress(),property, format, size, datastr);
}


/*
 * Pytputget Extended Method Implementation
 */
static PyObject* PyTine_tputgetEx (PyObject* pSelf, PyObject *args)
{
  //Arguments
  const char* address;
  const char* property;
  const char* format;
  char* datastr;
  int size = 0;
  PyObject *pList = NULL;

  //Extract the PyThon arguments
  if (!PyArg_ParseTuple(args,"sssO", &address, &property, &format, &pList))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
    return NULL;
  }

  //Check that the fifth parameter is a list
  if(!PyList_Check(pList))
  {
    PyErr_SetString(PyExc_TypeError,"pyput() fifth parameter must be a list");
    return NULL;
  }

  size = PyList_Size(pList);

  //datastr = pythonListToString(pList);
  string data = "";
  PyObject * pString = NULL;
  string tempString = "";
  stringstream str;
  long auxLong;
  double auxDouble;

  //Extract the list an compose a char
  for (int i=0;i<size;i++)
  {
    pString = PyList_GetItem(pList, i);
  	//Check the data type
  	if (!PyString_Check(pString))
  	{
  	  	switch (ftoi((char *) format))
		{
			case CF_SHORT:
			case CF_LONG:
				auxLong = PyLong_AsLong(pString);
				str << auxLong;
				str >> tempString;
				break;
			case CF_DOUBLE:
			case CF_FLOAT:
				auxDouble = PyFloat_AsDouble(pString);
				str << auxDouble;
				str >> tempString;
				break;
			default:
				Py_DECREF(pList);
	  			PyErr_SetString(PyExc_TypeError,"pyput() wrong input type");
	  			return NULL;
		}
	}
	else tempString = PyString_AsString(pString);
   	data = data + tempString + " ";

  }
  datastr = (char *) data.c_str();

  return PyTine_tputgetMain(address,property, format, size, datastr);

}


static PyObject* PyTine_GetProperties (PyObject*, PyObject *args)
{
	int size = 0;
	int sts = 0;
	int nlen;
	int i;
	char s[65];
	PyObject *pList = NULL;

    if (initialized)
	{
		//Get the size
		sts = PyTine::instance()->tget(PyTine::instance()->getAddress(), "NPROPERTIES", 1, CF_INTEGER, blob, PyTine::instance()->getTimeout());
		if (sts == 0 )
		{
			size = sval[0];
			sts = PyTine::instance()->tget(PyTine::instance()->getAddress(), "PROPERTIES", 37, CF_NAME32, blob, PyTine::instance()->getTimeout());
			if (sts == 0 )
			{
				//Creates the PyThon list to return
				//Py_DECREF(pList);
				pList = PyList_New(size);
				//Py_INCREF(Py_None);
				assert(PyList_Check(pList));
				//Build a list with the properties
				nlen = fmtsizeof(CF_NAME32);
				for (i=0; i<size; i++)
				{
					strncpy(s,cval + i*nlen,nlen);
					PyList_SetItem(pList, i, Py_BuildValue("s", s));
				}
				//Free the memory
				return Py_BuildValue("O", pList);
			}
		}
	}

    //PyObject *
    //Py_DECREF(pList);
    pList = PyList_New(1);
    assert(PyList_Check(pList));
    //Py_INCREF(pList);
    PyList_SetItem(pList, 0, Py_BuildValue("s", "PyTine Not Initialized"));
    return Py_BuildValue("O", pList);
    //return Py_BuildValue("s", "error");
}

static PyObject* PyTine_GetSrvAddress (PyObject*, PyObject *args)
{
	int sts = 0;
	int nlen;
	int i;
	char s[65];
	PyObject *pList = NULL;

	if (initialized)
	{
		sts = PyTine::instance()->tget(PyTine::instance()->getAddress(), "SRVADDR", 5, CF_NAME32, blob, PyTine::instance()->getTimeout());
		if (sts == 0 )
		{
			//Creates the PyThon list to return
			//Py_DECREF(pList);
			pList = PyList_New(5);
			//Py_INCREF(Py_None);
			assert(PyList_Check(pList));
			//Build a list with the properties
			nlen = fmtsizeof(CF_NAME32);
			for (i=0; i<5; i++)
			{
				strncpy(s,cval + i*nlen,nlen);
				PyList_SetItem(pList, i, Py_BuildValue("s", s));
			}
			//Free the memory??
			return Py_BuildValue("O", pList);
		}
	}

	pList = PyList_New(1);
	assert(PyList_Check(pList));
	PyList_SetItem(pList, 0, Py_BuildValue("s", "PyTine Not Initialized"));
	return Py_BuildValue("O", pList);
}

static PyObject* PyTine_GetNumProperties (PyObject*, PyObject *args)
{
	int size = 0;
	int sts = 0;

    if (initialized)
	{
		//Get the size
		sts = PyTine::instance()->tget(PyTine::instance()->getAddress(), "NPROPERTIES", 1, CF_INTEGER, blob, PyTine::instance()->getTimeout());
		if (sts == 0 )
		{
			size = sval[0];
			return Py_BuildValue("i", size);
		}
	}
    return Py_BuildValue("i", -1);
}

static PyObject* PyTine_GetStructFormatMain (char * tag)
{
	char errstr[256];
	int i = 0;
	int nlen = 0;
	int fmt = 0;
	int sts = 0;
	NAME64DBLDBL inputStruct;
	int size = 0;


	//Extract the format
	fmt = CF_NAME16;
	size = 1;

	//Call the C code and returns a Python object
	sts = PyTine::instance()->tputgetStruct(PyTine::instance()->getAddress(),"STRUCTFORMAT",size,fmt,blob_struct,tag);
	//No error in the TINE call
	if (sts == 0 )
	{

	  //Get the format and the size of the output from the last TINE call
	  fmt = GetCompletionDataType(-1);
	  size = GetCompletionDataSize(-1);

	  //Creates the PyThon list to return
	  PyObject *pList = PyList_New(size);
	  assert(PyList_Check(pList));

	  switch(fmt)
	  {
		case CF_NAME64DBLDBL:
		  nlen = fmtsizeof(CF_NAME64DBLDBL);
		  for (i=0; i<size; i++)
		  {
		    //Extract the NAME64
		    inputStruct = structFormat[i];
		  	PyList_SetItem(pList, i, Py_BuildValue("sff", inputStruct.name, inputStruct.d1val, inputStruct.d2val));
		  }
		  break;
		default:
		  PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported")); break;
		}
		return pList;

	  }
	  else
	  {
		  //Creates the PyThon list to return
		  PyObject *pList = PyList_New(1);
		  assert(PyList_Check(pList));
		  //Return the TINE Error
		  PyList_SetItem(pList, 0, Py_BuildValue("s", GetLastLinkError(sts,errstr)));
		  return pList;
	  }
}

static PyObject* PyTine_GetStructFormat (PyObject*, PyObject *args)
{
	char * tag;

	//Extract the PyThon arguments
	if (!PyArg_ParseTuple(args,"s", &tag))
	{
	    PyErr_SetString(PyExc_TypeError, "invalid parameter");
	  	return NULL;
	}
	return PyTine_GetStructFormatMain(tag);
}

/****************************************************************************************
 * CALLBACK IMPLEMENTATION FOR PYTHON
 ****************************************************************************************/
/*
 * Calls the PyThon function when receives a TINE event
 * Open question: how a send the data back to Python
 * 1) Identify the data type and size
 * 2) Support for more than one callback function
 * 3) Do we need an array of handlers? YES
 */
void Route_Event(int id, int cc, PyObject *list)
{
  PyObject *args, *pres;
  /* call Python handler */
  //Add to return the data
  args = Py_BuildValue("(iiO)", id, cc, list);   	  /* make arg-list */
  if (Handler == NULL)
  {
	 cout << "handler not registered" << endl;
  }
  else
  {
	 pres = PyEval_CallObject(Handler, args);      /* apply: run a call */
	 Py_DECREF(args);                              /* add error checks */
	 if (pres != NULL)
	 {
	   Py_DECREF(pres);
	 }
  }


}

void Route_EventEx(int id, int cc, PyObject *list)
{
  PyObject *args, *pres;
  //Add to return the data
  args = Py_BuildValue("(iiO)", id, cc, list);   	     /* make arg-list */
  pres = PyEval_CallObject(handlerArray[id], args);      /* apply: run a call */

  Py_DECREF(args);                                       /* add error checks */
  if (pres != NULL)
  {
    Py_DECREF(pres);
  }
}



/*
 * Callback function called from TINE
 */
void Trigger_TINEEx(int id, int cc)
{
	int size = 0 ;
	int fmt = 0;
	PyObject *args, *pres;

	size = GetCompletionDataSizeFromCallbackId(id);
	fmt = GetCompletionDataTypeFromCallbackId(id);

    //Extract the callback data
	DTYPE dout;
	dout.dFormat = CF_NAME64I;
    dout.dArrayLength = 1;
    UINT16 * status = new UINT16();
	GetDataFromCallbackId(id, &dout, status) ;

	if (cc == 0 )
	{
		int i = 0;
		int nlen = 0;
		char s[65];
		PyObject *pList = PyList_New(size);
		//assert(PyList_Check(pList)); //cause problems, it is needed
		switch (fmt)
		{
			case CF_LONG:
				for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("l", dout.data.lptr[i])); break;
			case CF_SHORT:
				for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("i", dout.data.sptr[i])); break;
			case CF_FLOAT:
				for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("f", dout.data.fptr[i])); break;
    	  	case CF_DOUBLE:
      	        for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("d", dout.data.dptr[i])); break;
    	  	case CF_TEXT:
    	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("s", dout.data.cptr[i])); break;
    	  	case CF_NAME16:
    	  	case CF_NAME32:
    	  	case CF_NAME48:
    	  	case CF_NAME64:
    	  		nlen = fmtsizeof(fmt);
    	  		for (i=0; i<size; i++)
    	  		{
    	  			strncpy(s,dout.data.cptr + i*nlen,nlen);
    	  			PyList_SetItem(pList, i, Py_BuildValue("s", s));
    	  		}
    	  		PyList_SetItem(pList, i, Py_BuildValue("s", "empty"));
    	  		break;
    	  	case CF_STRUCT:
    	  		//Resize the list to 1
    	  		PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported, use pygetStruct")); break;
    	  	default:
    	  		//Resize the list to 1
    	  		PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported")); break;
		}
		//Route_EventEx(id, cc, pList);
		//Py_INCREF(Py_None);

		//Add to return the data
		args = Py_BuildValue("(iiO)", id, cc, pList);
		pres = PyEval_CallObject(handlerArray[id], args);
		Py_DECREF(args);
		if (pres != NULL)
		{
		  Py_DECREF(pres);
		}
	}
    else
    {
    	char errstr[256];
    	//Creates the PyThon list to return
    	PyObject *pList = PyList_New(1);
    	//assert(PyList_Check(pList));
    	//Return the TINE Error
    	PyList_SetItem(pList, 0, Py_BuildValue("s", GetLastLinkError(cc,errstr)));

    	Route_EventEx(id, cc, pList);
    	Py_INCREF(Py_None);

    	//Free the memory
    	//Py_DECREF(pList);
    	//free(pList);
    }

}


/*
 * Callback function called from TINE
 */
void Trigger_TINE(int id, int cc)
{
	int size = 0 ;
	int fmt = 0;

	size = GetCompletionDataSizeFromCallbackId(id);
	fmt = GetCompletionDataTypeFromCallbackId(id);

	if (cc == 0 )
	{
		int i = 0;
		int nlen = 0;
		char s[65];
		PyObject *pList = PyList_New(size);
		//assert(PyList_Check(pList));
		switch (fmt)
		{
			case CF_LONG:
				for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("l", lval[i])); break;
			case CF_SHORT:
				for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("i", sval[i])); break;
			case CF_FLOAT:
				for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("f", fval[i])); break;
    	  	case CF_DOUBLE:
    	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("d", dval[i])); break;
    	  	case CF_TEXT:
    	  		for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("s", cval[i])); break;
    	  	case CF_NAME16:
    	  	case CF_NAME32:
    	  	case CF_NAME48:
    	  	case CF_NAME64:
    	  		nlen = fmtsizeof(fmt);
    	  		for (i=0; i<size; i++)
    	  		{
    	  			strncpy(s,cval + i*nlen,nlen);
    	  			PyList_SetItem(pList, i, Py_BuildValue("s", s));
    	  		}
    	  		break;
    	  	case CF_STRUCT:
    	  		//Resize the list to 1
    	  		PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported, use pygetStruct")); break;
    	  	default:
    	  		//Resize the list to 1
    	  		PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported")); break;
		}
		Route_Event(id, cc, pList);
		Py_INCREF(Py_None);

	}
    else
    {
    	char errstr[256];
    	//Creates the PyThon list to return
    	PyObject *pList = PyList_New(1);
    	//assert(PyList_Check(pList));
    	//Return the TINE Error
    	PyList_SetItem(pList, 0, Py_BuildValue("s", GetLastLinkError(cc,errstr)));

    	Route_Event(id, cc, pList);
    	Py_INCREF(Py_None);

    	//Free the memory
    	//Py_DECREF(pList);
    	//free(pList);
    }

	// Route the event to the Handler
	//Route_Event(id, cc, pList);
    //Py_INCREF(Py_None);
}


/* Callback Extended function
 * Automatically registers the handlers in an array
 */
static PyObject* PyTine_callbackEx (PyObject* self, PyObject *args)
{
  PyObject *result = NULL;
  PyObject *auxHandler = NULL;
  char * property;
  const char * format;
  char * mode;
  int size = 0;
  int sts = 0;
  int fmt = 0;

  if (handler_list_length == HANDLER_ARRAY_SIZE)
  {
	  PyErr_SetString(PyExc_TypeError, "List of Handlers already full (limit to 100)");
	  return NULL;
  }

  //Py_XINCREF(auxHandler);
  //Parse the input parameters
  if (!PyArg_ParseTuple(args,"sissO", &property, &size, &format, &mode, &auxHandler))
  {
	  PyErr_SetString(PyExc_TypeError, "invalid parameter");
	  return NULL;
  }

  //Register the function in the Array Handler
  Py_XDECREF(handlerArray[handler_list_length]);
  handlerArray[handler_list_length] = auxHandler;
  Py_XINCREF(handlerArray[handler_list_length]);
  //Py_XDECREF(auxHandler);

  //Extract the format
  fmt = ftoi((char *) format);
  //Subscribes to TINE
  cout << "Subscribing callback with id " << handler_list_length << endl;

  //Reserves memory for the callback data (how and when we delete this buffer)
  blob_async = new double[MAXIMUM_SIZE_ASYNC];

  //Subscribes in TINE with AttachLink
  sts = PyTine::instance()->tgetAsyncEx(PyTine::instance()->getAddress(), property, size, fmt, blob_async, Trigger_TINEEx, PyTine::instance()->getTimeout(), getAccessMode(mode), handler_list_length);

  //Increment the handler list
  handler_list_length++;

  Py_INCREF(Py_None);
  result = Py_None;
  return result;

}


/* Runs the TINE Channel */
static PyObject* PyTine_run (PyObject*, PyObject *args)
{
  for(;;)
  {
    SystemCycle(TRUE);
    sleep(1);
  }
  return Py_BuildValue("i", 0);
}

/* Runs the TINE Channel */
static PyObject* PyTine_SystemCycle (PyObject*, PyObject *args)
{
  SystemCycle(TRUE);
  return Py_BuildValue("i", 0);
}


/*********************************************************************************************
*  TINE FUNCTIONS
*********************************************************************************************/

static PyObject* PyTine_CloseGlobalLink(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int linkid = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &linkid))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = CloseGlobalLink(linkid);

  return (Py_BuildValue("i", result));

}

static PyObject* PyTine_CloseNetGlobal(PyObject* pSelf, PyObject *args)
{
  //Arguments
  const char * keyword;
  int result = 0;

  if (!PyArg_ParseTuple(args,"s", &keyword))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = CloseNetGlobal(keyword);

  return (Py_BuildValue("i", result));
}

PyObject* PyTine_FreeAccessLock(PyObject* pSelf, PyObject *args)
{
  //Arguments
  char * context;
  char * server;

  if (!PyArg_ParseTuple(args,"ss", &context, &server))
    PyErr_SetString(PyExc_TypeError, "invalid parameter");

  FreeAccessLock(context,server);

  return (Py_BuildValue("i", 0));
}

PyObject* PyTine_GetAccessLockInformation(PyObject* pSelf, PyObject *args)
{
  //Arguments
  char * context;
  char * server;
  char * name;
  char * ip;
  char * time;
  NAME32 * callerName = new NAME32();
  NAME32 * callerip   = new NAME32();
  NAME32 * timeLeft   = new NAME32();
  int result = 0;

  if (!PyArg_ParseTuple(args,"sssss", &context, &server, &name, &ip, &time))
    PyErr_SetString(PyExc_TypeError, "invalid parameter");

  strcpy(callerName->name,name);
  strcpy(callerip->name, ip) ;
  strcpy(timeLeft->name, time);

  printf("%s, %s\n", context, server);
  printf("%s,%s,%s\n", callerName->name, callerip->name, timeLeft->name);

  //NAME32 * test = new NAME32();
  //strcpy(test->name, callerName);

  result = GetAccessLockInformation(context,server, callerName,callerip,timeLeft);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetAccessLockStatus(PyObject* pSelf, PyObject *args)
{
  //Arguments
  char * context;
  char * server;
  int result = 0;

  if (!PyArg_ParseTuple(args,"ss", &context, &server))
    PyErr_SetString(PyExc_TypeError, "invalid parameter");

  result = GetAccessLockStatus(context, server);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetAlwaysRetry(PyObject* pSelf, PyObject *args)
{
  return (Py_BuildValue("i", GetAlwaysRetry()));
}

PyObject*  PyTine_GetClnRecvQueueDepth(PyObject* pSelf, PyObject *args)
{
	return (Py_BuildValue("i", GetClnRecvQueueDepth()));
}

PyObject*  PyTine_GetCompletionDataSize(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCompletionDataSize(i);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetCompletionDataSizeFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCompletionDataSizeFromCallbackId(i);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetCompletionDataType(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCompletionDataType(i);

  return (Py_BuildValue("i", result));

}

PyObject*  PyTine_GetCompletionDataTypeFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCompletionDataTypeFromCallbackId(i);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetCompletionSource(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCompletionSource(i);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetCompletionSourceFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCompletionSourceFromCallbackId(i);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetCompletionStatus(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  char * result;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCompletionStatus(i);

  return (Py_BuildValue("s", result));
}

PyObject*  PyTine_GetCompletionStatusFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  char * result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCompletionStatusFromCallbackId(i);

  return (Py_BuildValue("s", result));
}

PyObject*  PyTine_GetConnectionTableCapacity(PyObject* pSelf, PyObject *args)
{
  return (Py_BuildValue("i", GetConnectionTableCapacity()) );
}

PyObject*  PyTine_GetCurrentDataStatus(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCurrentDataStatus(i);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetCurrentDataStatusFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCurrentDataStatusFromCallbackId(i);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetCurrentDataTimeStamp(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  double result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCurrentDataTimeStamp(i);

  return (Py_BuildValue("d", result));
}

PyObject*  PyTine_GetCurrentDataTimeStampFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  double result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCurrentDataTimeStampFromCallbackId(i);

  return (Py_BuildValue("d", result));
}

PyObject*  PyTine_GetCurrentLinkStatus(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCurrentLinkStatus(i);

  return (Py_BuildValue("i", result));
}

PyObject*  PyTine_GetCurrentLinkStatusFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetCurrentLinkStatusFromCallbackId(i);

  return (Py_BuildValue("i", result));
}

/*
 * TO BE IMPLEMENTED
 */
PyObject*  PyTine_GetDataFromLinkId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int linkId = 0;
  /*unsigned short status;
  int result = 0;
  int fmt = 0;
  DTYPE dout;
  int nlen, i;
  char s[65];
  char errstr[256];*/

  if (!PyArg_ParseTuple(args,"i", &linkId))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  //Calls the TINE property
  //result = GetDataFromLinkId(linkId, &dout, &status);

  /*if (result == 0 )
  {
    //Creates the PyThon list to return
	int size = dout.dArrayLength;
	//PyObject *pList= PyList_New(dout.dArrayLength);
 	//assert(PyList_Check(pList));

    fmt = ftoi((char *) dout.dFormat);

    cout << "Size " << size << endl;

    switch(fmt)
	{
	    case CF_LONG:
			//for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("l", dout.data.lptr[i]));
	    	break;
		case CF_SHORT:
			//for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("i", dout.data.sptr[i]));
			break;
		case CF_FLOAT:
			cout << "FLOAT" << endl;
			//for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("f", dout.data.fptr[i])); break;
			//for (i=0; i<size; i++) cout << dout.data.fptr[i] << endl; break;
		case CF_DOUBLE:
			//for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("d", dout.data.dptr[i]));
			break;
		case CF_TEXT:
			//for (i=0; i<size; i++) PyList_SetItem(pList, i, Py_BuildValue("s", dout.data.cptr[i]));
			break;
		case CF_NAME16:
		case CF_NAME32:
		case CF_NAME48:
		case CF_NAME64:
			nlen = fmtsizeof(fmt);
		  	for (i=0; i<size; i++)
		  	{
		  		strncpy(s,dout.data.cptr + i*nlen,nlen);
		  		PyList_SetItem(pList, i, Py_BuildValue("s", s));
		  	}
		  	break;
		 default:
			 //PyList_SetItem(pList, 0, Py_BuildValue("s", "type not supported"));
			 break;
	}
    //Py_DECREF(pList);
    //free(pList);
    //return (Py_BuildValue("O", pList));
  }
  //else
  //{
    //Creates the PyThon list to return
    //PyObject *pList = PyList_New(1);
    //assert(PyList_Check(pList));
    //Return the TINE Error
    //PyList_SetItem(pList, 0, Py_BuildValue("s", GetLastLinkError(status,errstr)));
    //Py_DECREF(pList);
    //free(pList);
    //return (Py_BuildValue("O", pList));
  //}*/

  //Return a List with the requested data

  return (Py_BuildValue("s", "not Implemented"));
}

/*
 * TO BE IMPLEMENTED
 */
PyObject*  PyTine_GetDataFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments

  /*if (!PyArg_ParseTuple(args,"iOi", &i, &dout, &status))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }*/

  //result = GetCurrentLinkStatus(i);
  return (Py_BuildValue("s", "not implemented"));
}

PyObject*  PyTine_GetDataStamp(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int linkId = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &linkId))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetDataStamp(linkId);
  return (Py_BuildValue("i", result));

}

static PyObject*  PyTine_GetGlobalLinkId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  char * keyword;
  int result = 0;

  if (!PyArg_ParseTuple(args,"s", &keyword))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetGlobalLinkId(keyword);

  return (Py_BuildValue("i", result));
}

static PyObject*  PyTine_GetGlobalLinkIdFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetGlobalLinkIdFromCallbackId(i);

  return (Py_BuildValue("i", result));
}

static PyObject*  PyTine_GetGlobalsHeartbeat(PyObject* pSelf, PyObject *args)
{
	return (Py_BuildValue("i", GetGlobalsHeartbeat()) );
}
static PyObject*  PyTine_GetGlobalsTableCapacity(PyObject* pSelf, PyObject *args)
{
	return (Py_BuildValue("i", GetGlobalsTableCapacity()) );
}

static PyObject*  PyTine_GetGlobalsTimeStamp(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  double result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetGlobalsTimeStamp(i);

  return (Py_BuildValue("d", result));
}

static PyObject*  PyTine_GetLastGlobalDataSize(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetLastGlobalDataSize(i);

  return (Py_BuildValue("i", result));
}

static PyObject*  PyTine_GetLastGlobalDataSizeFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetLastGlobalDataSizeFromCallbackId(i);

  return (Py_BuildValue("i", result));
}

static PyObject*  PyTine_GetLastLinkError(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int cc = 0;
  char errstr[256];
  char * result;

  if (!PyArg_ParseTuple(args,"i", &cc))
  {
      PyErr_SetString(PyExc_TypeError, "invalid parameter");
      return NULL;
  }

  result = GetLastLinkError(cc, errstr);

  return (Py_BuildValue("s", result));
}

static PyObject*  PyTine_GetLinkName(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  char * result;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetLinkName(i);

  return (Py_BuildValue("s", result));
}

static PyObject*  PyTine_GetLinkNameFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  char* result;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetLinkNameFromCallbackId(i);

  return (Py_BuildValue("s", result));
}

static PyObject*  PyTine_GetLinkQueueDepth(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetLinkQueueDepth(i);

  return (Py_BuildValue("i", result));
}

static PyObject*  PyTine_GetSynchronizationTolerance(PyObject* pSelf, PyObject *args)
{
	return (Py_BuildValue("i", GetSynchronizationTolerance()) );
}

static PyObject*  PyTine_GetSystemDataStamp(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetSystemDataStamp(i);

  return (Py_BuildValue("i", result));
}

static PyObject*  PyTine_GetSystemDataStampFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetSystemDataStampFromCallbackId(i);

  return (Py_BuildValue("i", result));
}

static PyObject*  PyTine_GetTransferFlag(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetTransferFlag(i);

  return (Py_BuildValue("i", result));
}

static PyObject*  PyTine_GetTransferFlagFromCallbackId(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = GetTransferFlagFromCallbackId(i);

  return (Py_BuildValue("i", result));
}

/*
 * TO BE IMPLEMENTED
 */
static PyObject*  PyTine_ModifyLinkAttributes(PyObject* pSelf, PyObject *args)
{
  return (Py_BuildValue("s", "not implemented"));
}

/*
 * TO BE IMPLEMENTED
 */
static PyObject*  PyTine_recvNetGlobal(PyObject* pSelf, PyObject *args)
{
  return (Py_BuildValue("s", "not implemented"));
}

/*
 * TO BE IMPLEMENTED
 */
static PyObject*  PyTine_recvNetGlobalEx(PyObject* pSelf, PyObject *args)
{
  return (Py_BuildValue("s", "not implemented"));
}

/*
 * TO BE IMPLEMENTED
 */
static PyObject*  PyTine_recvNetGlobalEx2(PyObject* pSelf, PyObject *args)
{
  return (Py_BuildValue("s", "not implemented"));
}

/*
 * TO BE IMPLEMENTED
 */
static PyObject*  PyTine_RegisterCycleTriggerFunction(PyObject* pSelf, PyObject *args)
{
  return (Py_BuildValue("s", "not implemented"));
}

static PyObject*  PyTine_SetAccessLock(PyObject* pSelf, PyObject *args)
{
  //Arguments
  char * context;
  char * server;
  int lockType = 0;
  int lockDuration = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"ssii", &context, &server, &lockType, &lockDuration))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = SetAccessLock(context, server, lockType, lockDuration);

  return (Py_BuildValue("i", result));
}


static PyObject*  PyTine_SetAlwaysRetry(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  SetAlwaysRetry(i);

  return (Py_BuildValue("i", 0));

}

static PyObject*  PyTine_SetAutoLinkWatchdogs(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  SetAutoLinkWatchdogs(i);

  return (Py_BuildValue("i", 0));

}

static PyObject*  PyTine_SetClnRecvQueueDepth(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  SetClnRecvQueueDepth(i);

  return (Py_BuildValue("i", 0));
}

static PyObject*  PyTine_SetConnectionTableCapacity(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  SetConnectionTableCapacity(i);

  return (Py_BuildValue("i", 0));
}

static PyObject*  PyTine_SetGlobalsHeartbeat(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  SetGlobalsHeartbeat(i);

  return (Py_BuildValue("i", 0));

}

static PyObject*  PyTine_SetGlobalsTableCapacity(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int i = 0;

  if (!PyArg_ParseTuple(args,"i", &i))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  SetGlobalsTableCapacity(i);

  return (Py_BuildValue("i", 0));
}

static PyObject*  PyTine_SetLinkQueueDepth(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int linkId = 0;
  int depth = 0;

  if (!PyArg_ParseTuple(args,"ii", &linkId, &depth))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  SetLinkQueueDepth(linkId, depth);

  return (Py_BuildValue("i", 0));
}

static PyObject*  PyTine_SetLinkWatchdogPollingInterval(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int value = 0;
  int result = 0;

  if (!PyArg_ParseTuple(args,"i", &value))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = SetLinkWatchdogPollingInterval(value);

  return (Py_BuildValue("i", result));

}


static PyObject*  PyTine_SetNotificationTolerance(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int linkId = 0;
  float toleranceAbs = 0.0;
  float tolerancePerc = 0.0;
  int result = 0;


  if (!PyArg_ParseTuple(args,"iff", &linkId, &toleranceAbs, &tolerancePerc))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  result = SetNotificationTolerance(linkId, toleranceAbs, tolerancePerc);

  return (Py_BuildValue("i", result));
}


static PyObject*  PyTine_SetSuppressHeartbeatNotification(PyObject* pSelf, PyObject *args)
{
  //Arguments
  int value = 0;

  if (!PyArg_ParseTuple(args,"i", &value))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  SetSuppressHeartbeatNotification(value);

  return (Py_BuildValue("i", 0));
}

static PyObject*  PyTine_SetSynchronizationTolerance(PyObject* pSelf, PyObject *args)
{
  //Arguments
  double value = 0;

  if (!PyArg_ParseTuple(args,"f", &value))
  {
    PyErr_SetString(PyExc_TypeError, "invalid parameter");
  	return NULL;
  }

  SetSynchronizationTolerance(value);

  return (Py_BuildValue("i", 0));
}

/*
 * TO BE IMPLEMENTED
 */
static PyObject*  PyTine_UnregisterCycleTriggerFunction(PyObject* pSelf, PyObject *args)
{
  return (Py_BuildValue("s", "not implemented"));
}

/*
 * Init function
 */
PyMODINIT_FUNC initPyTine (void)
{
  Py_InitModule4 ("PyTine",  			// name of the module
		  PyTine_methods, 				// table of methods
		  "Python bindings for TINE", 	// doc string
		  0,
		  PYTHON_API_VERSION);

  if (PyErr_Occurred())
    Py_FatalError("Unable to initialize PyTine module");
}








